/*======================================================================*/
/*                                                                      */
/* Module Name: qres.c                                                  */
/*                                                                      */
/*----------------------------------------------------------------------*/
/*                                                                      */
/* Description: This sample program uses the Data Access API to         */
/* obtain information about reservations.                               */
/*                                                                      */
/*======================================================================*/
static char qres_sccsid[] = "@(#) src/ll/samples/lldata_access/qres.c, ll.api, ll_rjup, rjup0809b 1.6 2/7/07";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include "llapi.h"

extern char *optarg;

#define  STR_NULL(ptr) (ptr? ptr : "")
#define  FORMAT_BOOLEAN(v) ( (v)? "True" : "False")

void usage(void) {
	printf("\nUsage: \n\nqres [-l] [-u<owner>] [-g<group>] [-h<host>] [-B<BgBP>] [-b<begin] [-e<end>] [<reservation list>]\n\n");
	printf("Examples:\n");
	printf("qres -l \t\t: display complete information of all reservations\n");
	printf("qres -u dave -g NIC \t: display information of all reservations owned by\t\t\t\tuser dave and LoadLever group NIC.\n");
	printf("qres -h c94n04 \t\t: display information of all reservations\t\t\t\t\treserving time on host c94n04.\n");
	printf("qres -B R000   \t\t: display information of all reservations\t\t\t\t\treserving time on Blue Gene base partition R000.\n");
	printf("qres c94n04.1.r c94n04.4.r : display information of all reservations listed.\n");
	printf("qres -b 0 -e 120 \t: display information of all reservations overlapping\t\t\t\tthe time period between now and 2 hours from now.\n");
	printf("begin and end are the number of minutes from the current time.\n");
}

void print_strings(char **ppc) {
	int i = 0;
	if (!ppc) return;
	while (ppc[i]) {
		printf(" %s", ppc[i]);
		i++;
	}
	printf("\n");
}

void print_list(char **ppc) {
	int i = 0;
	if (!ppc) return;
	while (ppc[i]) {
		printf("                                 %s\n", ppc[i]);
		i++;
	}
}

int free_strings(char **ppc) {
	char **p = ppc;
	while (*p) {
		free(*p);
		p++;
	}
	free (ppc);
	return 0;
}

/* String for Reservation State */
const char * reservation_state_string(int state_code) {
	switch (state_code) {
	case  RESERVATION_WAITING:
		return "WAITING";
		break;
	case RESERVATION_SETUP:
		return "SETUP";
		break;
	case RESERVATION_ACTIVE:
		return "ACTIVE";
		break;
	case RESERVATION_ACTIVE_SHARED:
		return "ACTIVE_SHARED";
		break;
	case RESERVATION_CANCEL:
		return "CANCEL";
		break;
	case RESERVATION_COMPLETE:
		return "COMPLETE";
		break;
	default:
		return "UNDEFINED_STATE";
	}
}

int main(int argc, char *argv[]) {
	int opt;
	int long_listing = 0;
	char **selected_reservations = NULL;
	int rc, obj_count, err_code;
	LL_element *queryObject, *a_reservation;

	char *reservation_id = NULL;
	char *reservation_owner = NULL;
	char *reservation_group = NULL;
	time_t reservation_start_time;
	int reservation_duration;
	int reservation_mode_shared;
	int reservation_mode_remove_on_idle;
	int reservation_state;
	time_t reservation_creation_time;
	char *last_modified_by = NULL;
	time_t reservation_modification_time;
	char **reserved_nodes = NULL;
	char **bound_steps = NULL;
	char **reservation_users = NULL;
	char **reservation_groups = NULL;
	int num_bg_cnodes = 0;
	int res_bg_connection = MESH;
	int *res_bg_shape = NULL;
	char **res_bg_bps = NULL;
	char *res_bg_partition = NULL;
	char *u_list[2];
	char *g_list[2];
	char *h_list[2];
	char *B_list[2];
	int uflag = 0;
	int gflag = 0;
	int hflag = 0;
	int Bflag = 0;
	int begin_time = -1, end_time = -1;
	int print_title = 1;

	/* Parsing command line arguments */
	while ( (opt = getopt(argc, argv, "lu:g:h:b:e:B:")) != EOF) {
		switch (opt) {
		case 'u':
			u_list[1] = NULL;
			u_list[0] = optarg;
			uflag = 1;
			break;
		case 'g':
			g_list[1] = NULL;
			g_list[0] = optarg;
			gflag = 1;
			break;
		case 'h':
			h_list[1] = NULL;
			h_list[0] = optarg;
			hflag = 1;
			break;
		case 'B':
			B_list[1] = NULL;
			B_list[0] = optarg;
			Bflag = 1;
			break;
		case 'l':
			long_listing = 1;
			break;
		case 'b':
			begin_time = atoi(optarg);
			break;
		case 'e':
			end_time = atoi(optarg);
			break;
		default:
			usage();
			return(-1);
		}
	}
	if (argv[optind] != NULL) {
		selected_reservations = &argv[optind];
	}

	/* Initialize the RESERVATIONS query */
	queryObject = ll_query(RESERVATIONS);
	if (!queryObject) {
		printf("ll_query(RESERVATIONS) returned NULL\n");
		return(-1);
	}

	/* Set query parameters */
	if (selected_reservations) {
		printf("Query reservations ");
		print_strings(selected_reservations);
		rc = ll_set_request(queryObject, QUERY_RESERVATION_ID, selected_reservations, ALL_DATA);
	} else if (uflag || gflag || hflag || Bflag) {
		printf("Query reservations meeting the following conditions:\n");
		if (uflag) {
			printf("\tOwner is %s\n", u_list[0]);
			rc = ll_set_request(queryObject, QUERY_USER, u_list, ALL_DATA);
		}
		if (gflag) {
			printf("\tLoadLeveler group is %s\n", g_list[0]);
			rc = ll_set_request(queryObject, QUERY_GROUP, g_list, ALL_DATA);
		}
		if (hflag) {
			printf("\tReserving time on %s\n", h_list[0]);
			rc = ll_set_request(queryObject, QUERY_HOST, h_list, ALL_DATA);
		}
		if (Bflag) {
			printf("\tReserving time on %s\n", B_list[0]);
			rc = ll_set_request(queryObject, QUERY_BG_BASE_PARTITION, B_list, ALL_DATA);
		}
	} else {
		printf("Query all reservations\n");
		rc = ll_set_request(queryObject, QUERY_ALL, NULL, ALL_DATA);
	}
	if (!selected_reservations && begin_time >= 0) {
		begin_time *= 60;
		begin_time += time(0);
		printf("\tEnding after %s", ctime((const time_t*)&begin_time));
	}
	if (!selected_reservations && end_time >= 0) {
		end_time *= 60;
		end_time += time(0);
		printf("\tStarting before %s", ctime((const time_t*)&end_time));
	}
	if (rc) {
		printf("ll_set_request returned %d\n", rc);
		return(-1);
	}

	/* Get the requested a_reservation from the central manager daemon */
	a_reservation = ll_get_objs(queryObject, LL_CM, NULL, &obj_count, &err_code);
	printf("Query returned %d reservations.\n", obj_count);
	if (err_code) {
		printf("ll_get_objs failed with err_code %d.\n", err_code);
		return(-1);
	}
	if (obj_count < 1) {
		return(0);
	}

	while (a_reservation) {
		rc = ll_get_data(a_reservation, LL_ReservationID, &reservation_id);
		rc = ll_get_data(a_reservation, LL_ReservationOwner, &reservation_owner);
		rc = ll_get_data(a_reservation, LL_ReservationStatus, &reservation_state);
		rc = ll_get_data(a_reservation, LL_ReservationStartTime, &reservation_start_time);
		rc = ll_get_data(a_reservation, LL_ReservationDuration, &reservation_duration);
		if (!selected_reservations && begin_time >= 0) {
			if ((reservation_start_time+reservation_duration) <= begin_time) {
				a_reservation = ll_next_obj(queryObject);
				continue; /* reservation ends before begin_time, no overlap */
			}
		}
		if (!selected_reservations && end_time >= 0) {
			if (reservation_start_time >= end_time) {
				a_reservation = ll_next_obj(queryObject);
				continue; /* reservation starts after end_time, no overlap */
			}
		}
		if (!long_listing) {
			/* Short format. Show key information associated with this reservation on 1 line. */
			if (print_title) {
				/* Display the banner of short output listing. */
				printf("Reservation ID\t\t\tOwner\tState\tStart Time\n");
				printf("--------------------------------------------------------------------------\n");
				print_title = 0; /* print title once only */
			}
			printf("%s\t%s\t%s\t%s", reservation_id, reservation_owner, reservation_state_string(reservation_state), ctime(&reservation_start_time));
			a_reservation = ll_next_obj(queryObject);
			continue;   /* Go to next reservation. */
		}

		/* Long format.  */
		rc = ll_get_data(a_reservation, LL_ReservationMachines, &reserved_nodes);
		rc = ll_get_data(a_reservation, LL_ReservationJobs, &bound_steps);
		rc = ll_get_data(a_reservation, LL_ReservationModeShared, &reservation_mode_shared);
		rc = ll_get_data(a_reservation, LL_ReservationModeRemoveOnIdle, &reservation_mode_remove_on_idle);
		rc = ll_get_data(a_reservation, LL_ReservationGroup, &reservation_group);
		rc = ll_get_data(a_reservation, LL_ReservationCreateTime, &reservation_creation_time);
		rc = ll_get_data(a_reservation, LL_ReservationModifiedBy, &last_modified_by);
		rc = ll_get_data(a_reservation, LL_ReservationModifyTime, &reservation_modification_time);
		rc = ll_get_data(a_reservation, LL_ReservationUsers, &reservation_users);
		rc = ll_get_data(a_reservation, LL_ReservationGroups, &reservation_groups);
		rc = ll_get_data(a_reservation, LL_ReservationBgCNodes, &num_bg_cnodes);
		rc = ll_get_data(a_reservation, LL_ReservationBgConnection, &res_bg_connection);
		rc = ll_get_data(a_reservation, LL_ReservationBgShape, &res_bg_shape);
		rc = ll_get_data(a_reservation, LL_ReservationBgBPs, &res_bg_bps);
		rc = ll_get_data(a_reservation, LL_ReservationBgPartition, &res_bg_partition);


		printf("\n================ Reservation %s ==================\n", reservation_id);
		printf("                 reservation_id: %s\n", reservation_id);
		printf("      reservation_creation_time: %s", ctime(&reservation_creation_time));
		printf("              reservation_owner: %s\n", reservation_owner);
		printf("              reservation_group: %s\n", reservation_group);
		printf("         reservation_start_time: %s", ctime(&reservation_start_time));
		printf("           reservation_duration: %d minutes\n", reservation_duration);
		printf("        reservation_mode_shared: %s\n", FORMAT_BOOLEAN(reservation_mode_shared));
		printf("reservation_mode_remove_on_idle: %s\n", FORMAT_BOOLEAN(reservation_mode_remove_on_idle));
		printf("              reservation_state: %s\n", reservation_state_string(reservation_state));
		printf("               last_modified_by: %s\n", last_modified_by);
		printf("  reservation_modification_time: %s", ctime(&reservation_modification_time));
		printf("              reservation_users:");
		print_strings(reservation_users);
		printf("             reservation_groups:");
		print_strings(reservation_groups);
		printf("                 reserved_nodes:\n");
		print_list(reserved_nodes);
		printf("                    bound_steps:\n");
		print_list(bound_steps);
		if (num_bg_cnodes > 0) {
			printf("                  num_bg_cnodes: %d\n", num_bg_cnodes);
			printf("              res_bg_connection: %s\n", res_bg_connection==TORUS?"TORUS":"MESH");
			printf("                   res_bg_shape: %dx%dx%d\n", res_bg_shape[0], res_bg_shape[1], res_bg_shape[2]);
			printf("                     res_bg_bps:\n");
			print_list(res_bg_bps);
			printf("               res_bg_partition: %s\n", res_bg_partition);
		}

		free(reservation_id);
		free(reservation_owner);
		free(reservation_group);
		free(last_modified_by);
		free_strings(reserved_nodes);
		free_strings(bound_steps);
		free_strings(reservation_users);
		free_strings(reservation_groups);
		free_strings(res_bg_bps);
		free(res_bg_shape);
		free(res_bg_partition);

		res_bg_partition = NULL;
		reservation_id = NULL;
		reservation_owner = NULL;
		reservation_group = NULL;
		last_modified_by = NULL;
		reserved_nodes = NULL;
		bound_steps = NULL;
		reservation_users = NULL;
		reservation_groups = NULL;

		/* get next reservation */
		a_reservation = ll_next_obj(queryObject);
	}

	ll_free_objs(queryObject);
	ll_deallocate(queryObject);

	return 0;
}
