
/* **************************************************************
 * The formatjob() function in this module formats the
 * information stored in an LL_job structure and 
 * sends the results to stdout. It is intended to be used 
 * in conjunction with the GetHistory() interface. See the
 * module pacct.c in this directory for an usage example.
 * ************************************************************ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>
#include "llapi.h"

#define FORMAT_NOTIFICATION_SIZE 5
#define STR_NULL(ptr) (ptr ? ptr : "(none)")

int formatstep(LL_job_step *);

void convert_rusage64(struct rusage64 *, double *, double *, double *);

static
char *Notifications[FORMAT_NOTIFICATION_SIZE] = {
	"Always", "Complete", "Error", "Never", "Start"
};

static
char  *Status[LL_MAX_STATUS+1] = {
	"Idle", "Starting", "Running", "Removed", "Completed", "Hold",
	"Deferred", "Submission Err", "Vacate", "Not Run", "Not Queued"
};


/*
 * formatjob
 * function - formats an LL_job structure.
 * input - pointer to LL_job structure.
 *
 */
int formatjob(LL_job *job)
{
	int i;

	if (!job) return 0;
	printf("================== Job step %s ============\n", STR_NULL(job->job_name));
	printf("%25s: %s\n", "Job Name", STR_NULL(job->job_name));
	printf("%25s: %d\n", "Structure Version", job->version_num);
	printf("%25s: %s\n", "Owner", STR_NULL(job->owner));
	printf("%25s: %s\n", "Group", STR_NULL(job->groupname));
	printf("%25s: %s\n", "Submitting host", STR_NULL(job->submit_host));
	printf("%25s: %d\n", "Submitting Userid", job->uid);
	printf("%25s: %d\n", "Submitting Groupid", job->gid);
	printf("%25s: %d\n", "Number of steps", job->steps);

	for ( i = 0; i < job->steps; i++) {
		formatstep(job->step_list[i]);
	}

} /* end format_job */

int formatstep(LL_job_step *step)
{
	int i;
	LL_MACH_USAGE64     *p_mach_usage64;
	LL_DISPATCH_USAGE64 *p_dispatch_usage64;
	LL_EVENT_USAGE64    *p_event_usage64;
	time_t  tmp_time;
	double starter_utime, starter_stime, starter_total;
	double step_utime, step_stime, step_total;

	if (!step) return 0;
	printf("%25s: %s\n", "Schedd Host", STR_NULL(step->id.from_host));
	printf("%25s: %d\n", "Job Number", step->id.cluster);
	printf("%25s: %d\n", "Step Number", step->id.proc);
	printf("%25s: %s\n", "Step Name",  STR_NULL(step->step_name));

	if (step->q_date > 0) {
		tmp_time = (time_t) step->q_date;
		printf("%25s: %s", "Queue Date", ctime(&tmp_time));
	}
	else {
		printf("%25s: %s\n", "Queue Date", "N/A");
	}

	if (step->dispatch_time > 0 ) {
		tmp_time = (time_t) step->dispatch_time;
		printf("%25s: %s", "Dispatch Time", ctime(&tmp_time));
	}
	else {
		printf("%25s: %s\n", "Dispatch Time", "N/A");
	}

	if (step->start_time > 0 ) {
		tmp_time = (time_t) step->start_time;
		printf( "%25s: %s", "Start Time", ctime(&tmp_time));
	}
	else {
		printf("%25s: %s\n", "Start Time", "N/A");
	}

	if(step->dependency) {
	  printf( "%25s: %s\n", "Job Step Dependency", step->dependency);
	}

	if ((step->status >= 0) && (step->status <= LL_MAX_STATUS)) {
		printf( "%25s: %s\n", "Status", Status[step->status]);
	}
	else {
		printf( "%25s: %s\n", "Status", "Unknown");
	}

	if((step->status == LL_COMPLETED) || (step->status == LL_REMOVED)) {
		if(step->completion_date == 0) {
			printf("%25s: %s\n", "Completion Date", "(not recorded)");
		} else {
			tmp_time = (time_t) step->completion_date;
			printf("%25s: %s", "Completion Date", ctime(&tmp_time));
		}
		printf("%25s: %d\n", "Completion Code", step->completion_code);
	} else {
		printf("%25s: %s\n", "Completion Date", "(not completed)");
	}

	printf("%25s: %d\n", "Start Count", step->start_count);
	printf("%25s: %d\n", "User Priority", step->prio);

	if ((step->notification >= 0) && (step->notification < FORMAT_NOTIFICATION_SIZE)) {
		printf("%25s: %s\n", "Notifications", Notifications[step->notification]);
	}
	else {
		printf("%25s: %s\n", "Notifications", "");
	}

	if(step->image_size) {
		printf("%25s: %d KB\n", "Virtual Image Size", step->image_size);
	} else {
		printf("%25s: %s\n", "Virtual Image Size", "(not recorded)");
	}

	if (step -> flags & LL_CHECKPOINT) {
		printf("%25s: %s\n", "Checkpoint", "yes");
	} else {
		printf("%25s: %s\n", "Checkpoint", "no");
	}

	if(step->start_date == 0) {
		printf("%25s: %s\n", "Hold Job Until", "(no hold)");
	} else {
		tmp_time = (time_t) step->start_date;
		printf("%25s: %s", "Hold Job Until", ctime(&tmp_time));
	}

	printf("%25s: %s\n", "Cmd", STR_NULL(step->cmd));
	if (step->args) {
	  printf("%25s: %s\n", "Args", step->args);
	}
	if (step->env) {
	  printf("%25s: %s\n", "Env", step->env);
	}
	printf("%25s: %s\n", "In", STR_NULL(step->in));
	printf("%25s: %s\n", "Out", STR_NULL(step->out));
	printf("%25s: %s\n", "Err", STR_NULL(step->err));
	printf("%25s: %s\n", "Initial Working Directory", STR_NULL(step->iwd));
	printf("%25s: %s\n", "Requirements", STR_NULL(step->requirements));
	if (step->preferences) {
	printf("%25s: %s\n", "Preferences", step->preferences);
	}

	if (step->flags & LL_STEP_PVM3) {
		printf("%25s: %s\n", "Job Type", "Pvm3");		
	} else if (step->flags & LL_STEP_PARALLEL) { 
		printf("%25s: %s\n", "Job Type", "General Parallel");		
	} else {
		printf("%25s: %s\n", "Job Type", "Serial");		
	}

	if (step->flags & (LL_STEP_PVM3 + LL_STEP_PARALLEL)) {
		printf("%25s: %d\n", "Min Processors", step->min_processors);		
		printf("%25s: %d\n", "Max Processors", step->max_processors);		
	}

	if (step->flags & LL_NODE_USAGE_NOT_SHARED) {
		printf("%25s: %s\n", "Node Usage", "not_shared"); 
	} else {
		printf("%25s: %s\n", "Node Usage", "shared"); 
	}

	if ((step->processor_list != NULL) && (step->processor_list[0] != 0)) {
		if (step->num_processors == 1) {
			printf("%25s: %s\n", "Allocated host name", step->processor_list[0]);
		} else { 
			printf("%25s: %d\n", "Allocated host count", step->num_processors);
			printf("%25s: %s\n", "Allocated host names", step->processor_list[0]);
			for (i = 1; i < step->num_processors; i++) {
				printf("%25s  %s\n", "", step->processor_list[i]);
			}
		}
	} else {
		printf("%25s: %s\n", "Allocated host name", "(none)");
	}

	printf("%25s: %s\n", "Notify User", STR_NULL(step->notify_user));
	printf("%25s: %s\n", "Shell", STR_NULL(step->shell));
	printf("%25s: %s\n", "Class", STR_NULL(step->stepclass));

	if(step->limits64.cpu_hard_limit >= 0) {
		printf("%25s: %lld seconds\n", "Cpu Hard Limit", step->limits64.cpu_hard_limit);
	}

	if(step->limits64.cpu_soft_limit >= 0) {
		printf("%25s: %lld seconds\n", "Cpu Soft Limit", step->limits64.cpu_soft_limit);
	}
		
	if(step->limits64.data_hard_limit >= 0) {
		printf("%25s: %lld bytes\n", "Data Hard Limit", step->limits64.data_hard_limit);
	}
	 
	if(step->limits64.data_soft_limit >= 0) {
		printf("%25s: %lld bytes\n", "Data Soft Limit", step->limits64.data_soft_limit);
	}
	
	if(step->limits64.as_hard_limit >= 0) {
		printf("%25s: %lld bytes\n", "As   Hard Limit", step->limits64.as_hard_limit);
	}

	if(step->limits64.as_soft_limit >= 0) {
		printf("%25s: %lld bytes\n", "As   Soft Limit", step->limits64.as_soft_limit);
	}

	if(step->limits64.nproc_hard_limit >= 0) {
		printf("%25s: %lld \n", "Nproc Hard Limit", step->limits64.nproc_hard_limit);
	}

	if(step->limits64.nproc_soft_limit >= 0) {
		printf("%25s: %lld \n", "Nproc Soft Limit", step->limits64.nproc_soft_limit);
	}

	if(step->limits64.memlock_hard_limit >= 0) {
		printf("%25s: %lld bytes\n", "Memlock Hard Limit", step->limits64.memlock_hard_limit);
	}

	if(step->limits64.memlock_soft_limit >= 0) {
		printf("%25s: %lld bytes\n", "Memlock Soft Limit", step->limits64.memlock_soft_limit);
	}

	if(step->limits64.locks_hard_limit >= 0) {
		printf("%25s: %lld \n", "Locks Hard Limit", step->limits64.locks_hard_limit);
	}

	if(step->limits64.locks_soft_limit >= 0) {
		printf("%25s: %lld \n", "Locks Soft Limit", step->limits64.locks_soft_limit);
	}

	if(step->limits64.nofile_hard_limit >= 0) {
		printf("%25s: %lld \n", "Nofile Hard Limit", step->limits64.nofile_hard_limit);
	}

	if(step->limits64.nofile_soft_limit >= 0) {
		printf("%25s: %lld \n", "Nofile Soft Limit", step->limits64.nofile_soft_limit);
	}

	if(step->limits64.core_hard_limit >= 0) {
		printf("%25s: %lld bytes\n", "Core Hard Limit", step->limits64.core_hard_limit);
	}

	if(step->limits64.core_soft_limit >= 0) {
		printf("%25s: %lld bytes\n", "Core Soft Limit", step->limits64.core_soft_limit);
	}
		
	if(step->limits64.file_hard_limit >= 0) {
		printf("%25s: %lld bytes\n", "File Hard Limit", step->limits64.file_hard_limit);
	}
		
	if(step->limits64.file_soft_limit >= 0) {
		printf("%25s: %lld bytes\n", "File Soft Limit", step->limits64.file_soft_limit);
	}
		
	if(step->limits64.stack_hard_limit >= 0) {
		printf("%25s: %lld bytes\n","Stack Hard Limit", step->limits64.stack_hard_limit);
	}

	if(step->limits64.stack_soft_limit >= 0) {
		printf("%25s: %lld bytes\n","Stack Soft Limit", step->limits64.stack_soft_limit);
	}
		
	if(step->limits64.rss_hard_limit >= 0) {
		printf("%25s: %lld bytes\n", "Rss Hard Limit", step->limits64.rss_hard_limit);
	}

	if(step->limits64.rss_soft_limit >= 0) {
		printf("%25s: %lld bytes\n", "Rss Soft Limit", step->limits64.rss_soft_limit);
	}
		
	if(step->limits.hard_cpu_step_limit >= 0) {
		printf("%25s: %d seconds\n", "Job Cpu Hard Limit", step->limits.hard_cpu_step_limit);
	}
		
	if(step->limits.soft_cpu_step_limit >= 0) {
		printf("%25s: %d seconds\n", "Job Cpu Soft Limit", step->limits.soft_cpu_step_limit);
	}
		
	if(step->limits.hard_wall_clock_limit >= 0) {
		printf("%25s: %d seconds\n", "Wall Clock Hard Limit", step->limits.hard_wall_clock_limit);
	}

	if(step->limits.soft_wall_clock_limit >= 0) {
		printf("%25s: %d seconds\n", "Wall Clock Soft Limit", step->limits.soft_wall_clock_limit);
	}

	printf("%25s: %d\n","Number", step->number);

	if(step->comment != NULL) 
	  printf("%25s: %s\n","Comment", step->comment);

	printf("%25s: %s\n","Account",
	    ((step->account_no) ? step->account_no : "(none)"));

	if(step->nqs_info.nqs_submit != NULL) { 
	  printf("%25s: %s\n","NQS Submit Queue", step->nqs_info.nqs_submit);
 	}
	if(step->nqs_info.nqs_query != NULL) { 
	  printf("%25s: %s\n","NQS Query Queues", step->nqs_info.nqs_query);
	}
	if(step->tracker != NULL) {
	  printf("%25s: %s\n", "Job tracking exit", step->tracker);
	}
	if(step->tracker_arg != NULL) {
	  printf("%25s: %s\n", "Job tracking arg", step->tracker_arg);
	}

	convert_rusage64(&(step->usage_info64.starter_rusage64), &starter_utime,
	                 &starter_stime, &starter_total);
	printf( "%25s: %.6f\n", "Starter User Time", starter_utime);
	printf( "%25s: %.6f\n", "Starter System Time", starter_stime);
	printf( "%25s: %.6f\n", "Starter Total Time", starter_total);

	convert_rusage64(&(step->usage_info64.step_rusage64), &step_utime,
	                 &step_stime, &step_total);
	printf( "%25s: %.6f\n", "Step User Time", step_utime);
	printf( "%25s: %.6f\n", "Step System Time", step_stime);
	printf( "%25s: %.6f\n", "Step Total Time", step_total);

	printf("--------> Detail Usage Report Starts <--------\n");

	p_mach_usage64 = step->usage_info64.mach_usage64;
	while (p_mach_usage64 != NULL) {
		printf("%25s: %s\n", "Running host", STR_NULL(p_mach_usage64->name));
		printf("%25s: %f\n", "Machine speed", p_mach_usage64->machine_speed);
		p_dispatch_usage64 = p_mach_usage64->dispatch_usage64;
		while (p_dispatch_usage64 != NULL) {
			p_event_usage64 = p_dispatch_usage64->event_usage64;
			while (p_event_usage64 != NULL) {
				if (p_event_usage64->event == LL_LOADL_EVENT) {
					printf("%25s: %s\n", "Event", "System");
				} else if (p_event_usage64->event == LL_INSTALLATION_EVENT) {
					printf("%25s: %s\n", "Event", "Installation Defined");
				} else {
					printf("%25s: %s\n", "Event", "Not Defined");
				}
				printf("%25s: %s\n", "Event name", STR_NULL(p_event_usage64->name));
				convert_rusage64(&(p_event_usage64->starter_rusage64), 
					&starter_utime, &starter_stime, &starter_total);
				printf( "%25s: %.6f\n", "Starter User Time", starter_utime);
				printf( "%25s: %.6f\n", "Starter System Time", starter_stime);
				printf( "%25s: %.6f\n", "Starter Total Time", starter_total);
				convert_rusage64(&(p_event_usage64->step_rusage64),
					&step_utime, &step_stime, &step_total);
				printf( "%25s: %.6f\n", "Step User Time", step_utime);
				printf( "%25s: %.6f\n", "Step System Time", step_stime);
				printf( "%25s: %.6f\n", "Step Total Time", step_total);
				p_event_usage64 = p_event_usage64->next;
			}
			p_dispatch_usage64 = p_dispatch_usage64->next;
		}
		p_mach_usage64 = p_mach_usage64->next;
	}
	printf("--------> Detail Usage Report Ends <--------\n");

} /* end format_step_long */

void 
convert_rusage64(struct rusage64 *p_rusage64, double *utime, double *stime, double *total)
{
	if (!p_rusage64) { *utime = 0.0; *stime = 0.0; *total = 0.0; return; }
	*utime = (double)p_rusage64->ru_utime.tv_sec +
	         1.0e-6*((double)p_rusage64->ru_utime.tv_usec);
	*stime = (double)p_rusage64->ru_stime.tv_sec +
	         1.0e-6*((double)p_rusage64->ru_stime.tv_usec);
	*total = *utime + *stime;
}
