
/* This sample program shows how the ll_get_nodes() and ll_get_jobs()
 * functions of the Query API can be used to obtain information relating 
 * to machines and jobs in a LoadLeveler cluster. It also shows how
 * the functions ll_start_job() and ll_terminate_job() of the WorkLoad 
 * Management API can be used to start and cancel jobs.
 * Notes:
 *   1. The ll_get_nodes() and ll_get_jobs() functions are old interfaces
 *      to LoadLeveler. They are provided for backward compatibility 
 *      reasons only. It is recommended that you use the Data Access API
 *      to obtain machines and jobs information if you are developing
 *      new applications that interact with LoadLeveler.
 *   2. To use the ll_start_job() function, the SCHEDULER_TYPE keyword
 *      of the global LoadLeveler configuration file must be set to API
 *      (SCHEDULER_TYPE = API).
 *   3. To run the test, 2 jobs must have been submitted and are in IDLE state.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "llapi.h"

void show_LL_node(LL_node *);
void show_LL_job(LL_job *);
void error_report(int);
#define  STR_NULL(ptr) (ptr ? ptr : "(null)")

main(int argc, char **argv)
{
  int   i, rtc;
  LL_get_nodes_info nodes_info;
  LL_get_jobs_info jobs_info;
  LL_start_job_info *start_info;
  LL_terminate_job_info *cancel_info;

  /* Get information relating to machines in the LoadLeveler cluster. */
  /* The ll_get_nodes API interface is deprecated and is provided for */
  /* backward compatibility only.  ll_query(MACHINES) should be used  */
  /* instead.                                                         */
  printf("Machines Information =====================================\n\n");
  if ((rtc = ll_get_nodes(&nodes_info)) != API_OK)
    {
      error_report(rtc);
    }
  else
    {
      printf("Number of Nodes = %d\n", nodes_info.numNodes);
      for (i=0; i < nodes_info.numNodes; i++)
        {
          printf("------------------------------------------------------\n");
          printf("Node No. = %d:\n", i);
          show_LL_node(nodes_info.NodeList[i]);
        }
    }

  /* Get information relating to Jobs in the LoadLeveler cluster.     */
  /* The ll_get_jobs API interface is deprecated and is provided for  */
  /* backward compatibility only.  ll_query(JOBS) should be used      */
  /* instead.                                                         */
  printf("Jobs Information ========================================\n\n");
  if ((rtc = ll_get_jobs(&jobs_info)) != API_OK)
    {
      error_report(rtc);
    }
  else
    {
      printf("Number of Jobs = %d\n", jobs_info.numJobs);
      for (i=0; i < jobs_info.numJobs; i++)
        {
          printf("------------------------------------------------------\n");
          printf("Job No. = %d:\n", i);
          show_LL_job(jobs_info.JobList[i]);
        }
    }

  /* Invoke the function ll_start_job() to start a job step. This     */
  /* operation is meaningful only when used with an external          */
  /* scheduler (SCHEDULER_TYPE = API).                                */

  printf("Start a Job Step ========================================\n\n");
  /* First, request the structure that will be populated with         */
  /* information about the job to be started                          */
  start_info = (LL_start_job_info *)malloc(sizeof(LL_start_job_info));
  if (!start_info)
    {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
    }

  /* Now allocate the machine list that will be passed to the start   */
  /* command.  This example assumes the job will be run on only one   */
  /* node so a list of size 2 (a final NULL entry is required to end  */
  /* the list) is needed.                                             */
  /*                                                                  */
  /* In production code you would probably query attributes about the */
  /* job to decide how many machines are needed.                      */
  start_info->nodeList = (char **)malloc(2*sizeof(char *));
  if (!start_info->nodeList)
    {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
    }

  /* populate the machine list and end with a NULL                    */
  start_info->nodeList[0] = strdup(nodes_info.NodeList[0]->nodename);
  start_info->nodeList[1] = NULL;
  printf("The node selected to run job is: %s\n",nodes_info.NodeList[0]->nodename);

  /* Start the 1st job step of the 1st job in the list received from  */
  /* ll_get_jobs()                                                    */
  start_info->version_num = jobs_info.version_num;
  start_info->StepId.cluster = jobs_info.JobList[1]->step_list[0]->id.cluster;
  start_info->StepId.proc = jobs_info.JobList[1]->step_list[0]->id.proc;
  start_info->StepId.from_host = strdup(jobs_info.JobList[1]->step_list[0]->id.from_host);

  printf("Start data:\n\tVersion: %d\n\tCluster: %d\n\tStep ID: %d\n\tFrom host: %s\n\n",
         start_info->version_num,
         start_info->StepId.cluster,
         start_info->StepId.proc,
         start_info->StepId.from_host);


  if ((rtc = ll_start_job(start_info)) != API_OK)
    {
      error_report(rtc);
    }
  else
    {
      printf("ll_start_job() invoked to start job step: %s.%d.%d on machine: %s.\n\n",
             start_info->StepId.from_host, start_info->StepId.cluster,
             start_info->StepId.proc,  start_info->nodeList[0]);
    }


  /* If there were more than one job in the list of jobs returned,    */
  /* cancel the second one                                            */
  printf("Cancel a Job Step =======================================\n\n");
  /* First, request a structure that will be populated with           */
  /* information about the job to be canceled                         */
  cancel_info = (LL_terminate_job_info *)malloc(sizeof(LL_terminate_job_info));
  if (!cancel_info)
    {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
    }

  cancel_info->version_num = jobs_info.version_num;
  cancel_info->StepId.cluster = jobs_info.JobList[0]->step_list[0]->id.cluster;
  cancel_info->StepId.proc = jobs_info.JobList[0]->step_list[0]->id.proc;
  cancel_info->StepId.from_host = strdup(jobs_info.JobList[0]->step_list[0]->id.from_host);
  cancel_info->msg = NULL;

  printf("Cancel data:\n\tVersion: %d\n\tCluster: %d\n\tStep ID: %d\n\tFrom host: %s\n\n",
         cancel_info->version_num,
         cancel_info->StepId.cluster,
         cancel_info->StepId.proc,
         cancel_info->StepId.from_host);


  if ((rtc = ll_terminate_job(cancel_info)) != API_OK)
    {
      error_report(rtc);
    }
  else
    {
      printf("ll_terminate_job() invoked to cancel job step: %s.%d.%d.\n\n",
             cancel_info->StepId.from_host, cancel_info->StepId.cluster,
             cancel_info->StepId.proc);
    }

  free(start_info->nodeList[0]);
  free(start_info->nodeList);
  free(start_info->StepId.from_host);
  free(start_info);
  free(cancel_info->StepId.from_host);
  free(cancel_info);

  ll_free_nodes(&nodes_info);
  ll_free_jobs(&jobs_info);
}

void show_LL_node(LL_node *llnode)
{
  int i;
  struct tm *curtime;
  time_t bintime;

  printf("Node name = %s\n", STR_NULL(llnode->nodename));
  printf("Version = %d\n",llnode->version_num);

  /* Print the time the config file was last read on the machine       */
  bintime = (time_t)llnode->configtimestamp;
  curtime = localtime(&bintime);
  printf("Config Timestamp = %s", asctime(curtime));

  /* Print the last time the state of this machine changed             */
  bintime = (time_t)llnode->time_stamp;
  curtime = localtime(&bintime);
  printf("Timestamp = %s", asctime(curtime));

  /* Other machine characteristics                                     */
  printf("Virtual Memory = %d KB\n",llnode->virtual_memory);
  printf("Virtual Memory (64-bit) = %lld KB\n",llnode->virtual_memory64);
  printf("Real Memory = %d MB\n",llnode->memory);
  printf("Real Memory (64-bit) = %lld MB\n",llnode->memory64);
  printf("Disk = %d KB\n",llnode->disk);
  printf("Disk (64-bit) = %lld KB\n",llnode->disk64);
  printf("LoadAvg = %f\n",llnode->loadavg);
  printf("Speed = %f\n",llnode->speed);
  printf("Pool = %d\n",llnode->pool);
  printf("Cpus = %d\n",llnode->cpus);
  /* How long has it been (in secs) since this machine was used        */
  /* interactively?                                                    */
  printf("KeyboardIdle = %d\n",llnode->keywordidle);
  printf("Arch = %s\n", STR_NULL(llnode->arch));
  printf("Opsys = %s\n", STR_NULL(llnode->opsys));
  printf("Node state = %s\n", STR_NULL(llnode->state));

  /* Print the maximum number of jobs that should be running           */
  /* concurrently on this machine                                      */
  printf("Node max_starters = %d\n",llnode->max_starters);

  /* How many jobs have been submitted to the schedd on this machine?  */
  printf("Total Jobs submitted = %d\n",llnode->totaljobs);

  /* Print the list of jobs currently running on this machine          */
  if (llnode->steplist[0].cluster != 0)
    {
      i = 0;
      printf("Job Steps on node: \n");
      while (llnode->steplist[i].cluster != 0)
        {
          printf("\tStep %d = %s.%d.%d\n",
                 i, llnode->steplist[i].from_host,
                 llnode->steplist[i].cluster,
                 llnode->steplist[i].proc);
          i++;
        }
    }
  else
    {
      printf("No Job Steps on node.\n");
    }

  /* Print names of all adapter devices configured on this machine     */
  if (llnode->adapter != NULL)
    {
      i = 0;
      printf("Adapters on node: \n");
      while (llnode->adapter[i] != NULL)
        printf("\tNode adapter = %s\n",llnode->adapter[i++]);
    }
  else
    {
      printf("Node adapter = NULL\n");
    }

  /* Print all features configured on this machine                     */
  if (llnode->feature != NULL)
    {
      i = 0;
      printf("Features on node: \n");
      while (llnode->feature[i] != NULL)
        printf("\tfeature = %s\n",llnode->feature[i++]);
    }
  else
    {
      printf("Node features = NULL\n");
    }

  /* Print all classes configured to run on this machine               */
  if (llnode->job_class != NULL)
    {
      i = 0;
      printf("Job classes on node: \n");
      while (llnode->job_class[i] != NULL)
        printf("\tJob class = %s\n",llnode->job_class[i++]);
    }
  else
    {
      printf("Node job_class = NULL\n");
    }
}

void show_LL_job(LL_job *lljob)
{
  int i;
  struct tm *curtime;
  time_t bintime;

  printf("Job name = %s\n", STR_NULL(lljob->job_name));
  printf("Number of job steps = %d:\n",lljob->steps);
  for (i=0; i < lljob->steps; i++)
    {
      printf("\nJob ID = %s.%d.%d.\n",
             lljob->step_list[i]->id.from_host,
             lljob->step_list[i]->id.cluster,
             lljob->step_list[i]->id.proc);

      if (lljob->step_list[i]->dispatch_time)
        {
          bintime = (time_t)lljob->step_list[i]->dispatch_time;
          curtime = localtime(&bintime);
          printf("Status %d, Dispatch time: %s",
                 lljob->step_list[i]->status, asctime(curtime));
        }
      else
        {
          printf("Status %d, Dispatch time: %s\n",
                 lljob->step_list[i]->status, "N/A");
        }

      printf("Step name = %s\n", STR_NULL(lljob->step_list[i]->step_name));
      printf("Class = %s\n", STR_NULL(lljob->step_list[i]->stepclass));
      printf("Hard Limits:\n");
      printf("\tCPU           = %d\n",lljob->step_list[i]->limits.cpu_hard_limit);
      printf("\tWallClock     = %d\n",lljob->step_list[i]->limits.hard_wall_clock_limit);
      printf("\tFile          = %d\n",lljob->step_list[i]->limits.file_hard_limit);
      printf("\tFile (64-bit) = %lld\n",lljob->step_list[i]->limits64.file_hard_limit);
      printf("Soft Limits:\n");
      printf("\tCPU           = %d\n",lljob->step_list[i]->limits.cpu_soft_limit);
      printf("\tWallClock     = %d\n",lljob->step_list[i]->limits.soft_wall_clock_limit);
      printf("\tFile          = %d\n",lljob->step_list[i]->limits.file_soft_limit);
      printf("\tFile (64-bit) = %lld\n",lljob->step_list[i]->limits64.file_soft_limit);
    }
}

void error_report(int rtc)
{
  switch (rtc)
    {
    case API_INVALID_INPUT:
      fprintf(stderr, "Invalid parameter.\n");
      break;

    case API_CANT_MALLOC:
      fprintf(stderr, "Out of memory.\n");
      break;

    case API_CANT_CONNECT:
      fprintf(stderr, "Can't connect to CM.\n");
      break;

    case API_CONFIG_ERR:
      fprintf(stderr, "Configuration error.\n");
      break;

    case API_CANT_TRANSMIT:
      fprintf(stderr, "Transaction problem.\n");
      break;

    case API_CANT_AUTH:
      fprintf(stderr, "Not an administrator.\n");
      break;

    case API_CANT_FIND_PROC:
      fprintf(stderr, "No such step in CM.\n");
      break;

    case API_WRNG_PROC_VERSION:
      fprintf(stderr, "Wrong proc version number.\n");
      break;

    case API_WRNG_PROC_STATE:
      fprintf(stderr, "Wrong step state to start.\n");
      break;

    case API_MACH_NOT_AVAIL:
      fprintf(stderr, "Machine not available.\n");
      break;

    case API_CANT_FIND_RUNCLASS:
      fprintf(stderr, "Run class not available.\n");
      break;

    case API_REQ_NOT_MET:
      fprintf(stderr, "Requirements not met.\n");
      break;

    case API_WRNG_MACH_NO:
      fprintf(stderr, "Wrong machine number.\n");
      break;

    case API_LL_SCH_ON:
      fprintf(stderr, "LL scheduler on.\n");
      break;

    case API_MACH_DUP:
      fprintf(stderr, "Duplicated machine names.\n");
      break;

    case API_NO_DCE_CRED:
      fprintf(stderr, "No dce credentials.\n");
      break;

    case API_INSUFFICIENT_DCE_CRED:
      fprintf(stderr, "DCE credential lifetime less than 300 seconds.\n");
      break;

    case API_64BIT_DCE_ERR:
      fprintf(stderr, "LoadLeveler 64-bit APIs not supported when DCE is enabled.\n");
      break;

    default:
      fprintf(stderr, "Unknown error.\n");
    }
}
